import browser from "webextension-polyfill";

// TODO: check if this is the correct way to calculate specificity. css parser might be better
function calculateSpecificity(selector: string): [number, number, number] {
    // 결과를 저장할 Specificity 배열: [ID, Class/Attribute/Pseudo-class, Tag/Pseudo-element]
    const specificity = [0, 0, 0] as [number, number, number];

    // 정규식을 사용하여 선택자에서 ID, 클래스/속성/가상 클래스, 태그/가상 요소를 분리
    const regex = {
        id: /#[\w-]+/g, // ID 선택자
        classAttrPseudo: /\.[\w-]+|\[[^\]]+\]|:[^\s>+~.#\[:]+/g, // 클래스, 속성 선택자, 가상 클래스
        tagPseudoElement: /^[a-z]+|::[\w-]+/gi // 태그 이름 또는 가상 요소 (::before, ::after 등)
    };

    // ID 선택자 개수 계산
    const idMatches = selector.match(regex.id);
    if (idMatches) {
        specificity[0] += idMatches.length;
    }

    // 클래스, 속성 선택자 및 가상 클래스 개수 계산
    const classAttrPseudoMatches = selector.match(regex.classAttrPseudo);
    if (classAttrPseudoMatches) {
        specificity[1] += classAttrPseudoMatches.length;
    }

    // 태그 및 가상 요소 개수 계산
    const tagPseudoElementMatches = selector.match(regex.tagPseudoElement);
    if (tagPseudoElementMatches) {
        specificity[2] += tagPseudoElementMatches.length;
    }

    return specificity;
}

async function getCssRules(sheet: CSSStyleSheet) {
    try {
        return sheet.cssRules;
    } catch (error) {
        return null;
    }
}

async function getCssRulesCORS(sheet: CSSStyleSheet) {
    const r = await getCssRules(sheet);
    if (r) {
        return r;
    }
    try {
        // TODO: refactor this to extract the fetch function
        const text = await browser.runtime.sendMessage({ type: "fetch", url: sheet.href, resType: "text" });
        console.log("Fetched CSS", text.slice(0, 100));
        const styleElem = document.createElement("style");
        styleElem.textContent = text;
        // TODO: check if this is the correct way.
        // 스타일을 넣지 않으면 sheet가 생성되지 않음. 
        // 브라우저 엔진 말고 파싱 라이브러리 사용해서 성능 테스트 필요
        document.head.appendChild(styleElem);
        const rules = styleElem.sheet?.cssRules;
        document.head.removeChild(styleElem);
        return rules;
    } catch (error) {
        console.error("Error fetching CSS", error);
        return null;
    }
}

class StyleStore {
    private store = new Map<string, { rule: CSSStyleRule, specificity: [number, number, number] }>();
    public add(rule: CSSStyleRule, order = 0) {
        this.store.set(rule.selectorText, {
            rule,
            specificity: calculateSpecificity(rule.selectorText)
        });
    }
    public get(selector: string) {
        return this.store.get(selector);
    }
    public getAll() {
        return this.store;
    }
    static async fromDocument() {
        const store = new StyleStore();
        const sheets = document.styleSheets;
        for (let i = 0; i < sheets.length; i++) {
            let sheet = sheets[i];
            let rules = await getCssRulesCORS(sheet as CSSStyleSheet);
            if (rules) {
                for (let j = 0; j < rules.length; j++) {
                    const rule = rules[j];
                    if (rule instanceof CSSStyleRule) {
                        const styleRule = rule;
                        store.add(styleRule);
                    }
                    else if (rule instanceof CSSMediaRule) {
                        const conditionText = rule.conditionText;
                        // assume all media rules are screen or bigger than 1024px
                        if (!conditionText || window.matchMedia(conditionText).matches) {
                            const mediaRules = rule.cssRules;
                            if (mediaRules) {
                                for (let k = 0; k < mediaRules.length; k++) {
                                    const mediaStyleRule = mediaRules[k];
                                    if (mediaStyleRule instanceof CSSStyleRule) {
                                        store.add(mediaStyleRule);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return store;
    }
}

let styleStore: StyleStore | null = null;
async function getStoreSingleton() {
    if (!styleStore) {
        styleStore = await StyleStore.fromDocument();
    }
    return styleStore;
}

async function getAllAppliedRuleOfElement(element: Element) {
    const store = await getStoreSingleton();
    const rules = [] as { rule: CSSStyleRule, specificity: [number, number, number], selectorText: string }[];
    for (const [selector, { rule, specificity }] of store.getAll()) {
        if (element.matches(selector)) {
            rules.push({
                rule,
                specificity,
                selectorText: selector
            });
        }
    }
    rules.sort((a, b) => {
        if (a.specificity[0] !== b.specificity[0]) {
            return a.specificity[0] - b.specificity[0];
        } else if (a.specificity[1] !== b.specificity[1]) {
            return a.specificity[1] - b.specificity[1];
        } else {
            return a.specificity[2] - b.specificity[2];
        }
    });
    return rules;
}

export type VElementNode = {
    type: "element";
    tagName: string;
    style: { [key: string]: string };
    children: VNode[];
}

export type VTextNode = {
    type: "text";
    text: string;
}

export type VNode = VElementNode | VTextNode;

export function styleFromCSSStyleRule(rule: CSSStyleRule) {
    const style = {} as { [key: string]: string };
    for (let i = 0; i < rule.style.length; i++) {
        const property = rule.style.item(i);
        style[property] = rule.style.getPropertyValue(property);
    }
    return style;
}

import { cssPropertiesDefault } from "./CSSPropertiesDefaultValue";

export async function cloneWithStyle(element: HTMLElement): Promise<VNode> {
    const dom = {
        type: "element",
        tagName: element.tagName.toLowerCase(),
        style: {} as { [key: string]: string },
        children: [] as VNode[]
    } as VElementNode;
    const rules = await getAllAppliedRuleOfElement(element);
    // rules.reverse();
    // const computedStyle = getComputedStyle(element);
    const cssInheritanceSet = new Set(cssInheritanceProperties);
    for (const { rule } of rules) {
        const style = styleFromCSSStyleRule(rule);
        for (const property in style) {
            const value = style[property];
            if (value === "initial" && (!cssInheritanceSet.has(property))) {
                continue;
            }
            if (cssPropertiesDefault[property as keyof typeof cssPropertiesDefault] === value && (!cssInheritanceSet.has(property))) {
                continue;
            }
            dom.style[property] = style[property];
        }
    }
    element.style.cssText.split(";")
    .filter((property) => property.trim())
    .forEach((property) => {
        const [key, value] = property.split(":");
        dom.style[key.trim()] = value.trim();
    });
    for (let i = 0; i < element.childNodes.length; i++) {
        const child = element.childNodes[i];
        if (child.nodeType === Node.ELEMENT_NODE) {
            dom.children.push(await cloneWithStyle(child as HTMLElement));
        }
        else if (child.nodeType === Node.TEXT_NODE) {
            const text = child.textContent?.trim();
            if (text) {
                dom.children.push({
                    type: "text",
                    text
                });
            }
        }
    }
    return dom;
}

import { generatedProperties } from "./supportedCSSProperties";

const cssInheritanceProperties = (() => {
    const inherit = generatedProperties.filter(p => p.inherited)
    const longhands = new Set<string>();
    for (const property of inherit) {
        if (property.longhands) {
            property.longhands.forEach(longhand => longhands.add(longhand));
        }
    }
    return inherit.filter(p => !longhands.has(p.name))
    .map(p => p.name)
    .filter(p => !p.startsWith("-webkit-"));
    // TODO: remove this filter after fixing the issue with -webkit- properties
})();

export function getInheritedStyle(element: HTMLElement) {
    const style = {} as { [key: string]: string };
    const computedStyle = window.getComputedStyle(element);
    for (const property of cssInheritanceProperties) {
        style[property] = computedStyle.getPropertyValue(property);
    }
    return style;
}


// function getComputedStyleWithoutDefault(element) {
// const inheritedProperties = new Set([
//     "azimuth",
//     "border-collapse",
//     "border-spacing",
//     "caption-side",
//     "color",
//     "cursor",
//     "direction",
//     "elevation",
//     "empty-cells",
//     "font-family",
//     "font-size",
//     "font-style",
//     "font-variant",
//     "font-weight",
//     "font",
//     "letter-spacing",
//     "line-height",
//     "list-style-image",
//     "list-style-position",
//     "list-style-type",
//     "list-style",
//     "orphans",
//     "pitch-range",
//     "pitch",
//     "quotes",
//     "richness",
//     "speak-header",
//     "speak-numeral",
//     "speak-punctuation",
//     "speak",
//     "speech-rate",
//     "stress",
//     "text-align",
//     "text-indent",
//     "text-transform",
//     "visibility",
//     "voice-family",
//     "volume",
//     "white-space",
//     "widows",
//     "word-spacing",
// ]);
//     const computedStyle = window.getComputedStyle(element);
//     const parentComputedStyle = window.getComputedStyle(element.parentElement);
//     const style = {};
//     // TODO: add element's style to style
//     const defaultStyle = window.getComputedStyle(document.documentElement);
//     for (let i = 0; i < computedStyle.length; i++) {
//         const property = computedStyle[i];
//         if (inheritedProperties.has(property)) {
//             if (computedStyle[property] !== parentComputedStyle[property]) {
//                 style[property] = computedStyle[property];
//             }
//         } else {
//             // if default value is not equal to computed value then add it to style
//             // remove default value like 0px, 0em, 0%, none, normal, auto
//             const value = computedStyle.getPropertyValue(property);
//             if (value !== defaultStyle.getPropertyValue(property)) {
//                 style[property] = value;
//             }
//         }
//     }
//     return style;
// }